from requests import get
from lunespy.server import Node


def all_node_conected_in_node_url(node_url: str = None) -> dict:
    if node_url == None:
        full_url = f'{Node.mainnet_url.value}/peers/all'
    else:
        full_url = f'https://{node_url}/peers/all'
    
    response = get(full_url)

    if response.status_code in range(200, 300):
        return {
            'status': 'ok',
            'response': response.json()
        }
    else:
        return {
            'status': 'error',
            'response': response.text
        }


def node_version(node_url: str = None) -> dict:
    if node_url == None:
        full_url = f'{Node.mainnet_url.value}/utils/lunesnode/version'
    else:
        full_url = f'https://{node_url}/utils/lunesnode/version'
    
    response = get(full_url)

    if response.status_code in range(200, 300):
        return {
            'status': 'ok',
            'response': response.json()
        }
    else:
        return {
            'status': 'error',
            'response': response.text
        }


def version_all_lunes_node_conected(node_url: str = None) -> dict:
    if node_url == None:
        full_url = f'{Node.mainnet_url.value}/peers/connected'
    else:
        full_url = f'https://{node_url}/peers/connected'
    
    response = get(full_url)

    if response.status_code in range(200, 300):
        return {
            'status': 'ok',
            'response': [
                {
                    'node_url': data['address'][1:],
                    'version': data['applicationVersion']
                }
                for data in response.json()['peers']
            ]
        }
    else:
        return {
            'status': 'error',
            'response': response.text
        }



class bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKCYAN = '\033[96m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


def log_data(data: dict) -> None:
    list(map(
        lambda tuple: print(
            f"{tuple[0]}{bcolors.OKGREEN} ─> {str(tuple[1])}{bcolors.ENDC}"
        ),
        data.items()
        )
    )


def export_json(data: dict, name: str, path: str) -> bool:
    import json
    path = path.replace("//", "/")
    full_path = f"{path}{name}.json"
    try:
        with open(full_path, 'w') as file:
            file.write( json.dumps(data) )
    except Exception as msg:
        raise Exception(
            bcolors.FAIL + f"[Error] File Don't Saved Because:\n{msg}" + bcolors.ENDC
        )

    return f"file save in {full_path}"


def generate_log() -> None:
    from subprocess import check_output

    def get_logs() -> list:
        return check_output(
            'git log --pretty="- [%h](%H) %s [%ai]"',
            shell=True
        ).decode().split('\n')
    
    def logs_to_changelog(logs: list) -> list:
        range_date = {}
        for line in logs:
            range_date[line[-27:-17]] = []
            for commit in logs:
                if line[-27:-17] == commit[-27:-17]:
                    range_date[line[-27:-17]].append(commit)

        changelog = ['# Changelog\n']
        for date in range_date.keys():
            changelog.append(f"\n## {date}\n")
            for commit in range_date[date]:
                edited_commit = commit[:-29] + '\n'
                changelog.append(edited_commit)
        
        return changelog

    def save_changelog(changelog: list) -> None:
        with open('./CHANGELOG.md', 'w') as file:
            file.writelines(changelog)

    save_changelog(
        logs_to_changelog(
            get_logs()
            )
    )


def now() -> int:
    from time import time

    return int(
        time() * 1000
    )


def lunes_to_unes(lunes: float or int) -> int:
    return int(lunes * 10e7)


def unes_to_lunes(unes: int) -> float:
    return float(unes / 10e7)


def sha256(object: object) -> str:
    from hashlib import sha256

    return sha256(
        str(object).encode()
    ).hexdigest()


def drop_none(data: dict) -> dict:
    validate_keys = list(filter(
        lambda key: data[key] != None,
        data.keys()
    ))

    return {
        key: data[key]
        for key in data.keys()
        if key in validate_keys
    }



from requests import get


def transaction_from_id(id: str) -> dict:
    full_url = f'http://lunesnode-testnet.lunes.io/transactions/info/{id}'
    response = get(full_url)

    if response.status_code in range(200, 300):
        return {
            'status': 'ok',
            'response': response.json()
        }
    else: 
        return {
            'status': 'error',
            'response': response.text
        }


def unconfirmed_transaction(node_url: str) -> dict:
    full_url = f'https://{node_url}/transactions/unconfirmed'
    response = get(full_url)

    if response.status_code in range(200, 300):
        return {
            'status': 'ok',
            'response': response.json()
        }
    else:
        return {
            'stauts': 'error',
            'response': response.text
        }


def transactions_from_address(node_url: str, address: str, limit_transactions: int) -> dict:
    full_url = f'https://{node_url}/transactions/address/{address}/limit/{limit_transactions}'
    response = get(full_url)

    if response.status_code in range(200, 300):
        return {
            'status': 'ok',
            'response': response.json()
        }
    else: 
        return {
            'status': 'error',
            'response': response.text
        }



from lunespy.client.transactions import BaseTransaction
from lunespy.client.wallet import Account
from lunespy.utils import drop_none

class TransferToken(BaseTransaction):
    def __init__(self, sender: Account, receiver: Account, asset_fee:int = None,
                 asset_id = None, amount:float = None, timestamp: int = None,
                 fee: int = None) -> None:
        self.transfer_data: dict = drop_none({
            'timestamp': timestamp,
            'asset_fee': asset_fee,
            'asset_id': asset_id,
            'amount': amount,
            'fee': fee
        })
        super().__init__('Transfer', self.transfer_data)
        self.sender: Account = sender
        self.receiver: Account = receiver
        self.history: list = []


    @property
    def ready(self) -> bool:
        from lunespy.client.transactions.transfer.validators import validate_transfer

        return validate_transfer(self.sender, self.receiver, self.transfer_data)


    @property
    def transaction(self) -> dict:
        from lunespy.client.transactions.transfer.validators import mount_transfer

        return super().transaction(
            mount_tx=mount_transfer,
            sender=self.sender,
            receiver=self.receiver,
            transfer_data=self.transfer_data)


    def send(self, node_url: str = None) -> dict:
        from lunespy.client.transactions.transfer.validators import send_transfer

        tx = super().send(send_transfer, node_url)
        self.history.append(tx)
        return tx



from enum import Enum

class Node(Enum):
    mainnet_url: str = 'https://lunesnode.lunes.io'
    testnet_url: str = 'https://lunesnode-testnet.lunes.io'
    mainnet_blockexplorer: str = 'https://blockexplorer.lunes.io'
    testnet_blockexplorer: str = 'https://blockexplorer-testnet.lunes.io'
    mainnet_total_supply: float = 150_728_537.61498705
    testnet_total_supply: float = 800_100_000.00000000  


from lunespy.client.wallet import Account


def mount_transfer(sender: Account, receiver: Account, transfer_data: dict) -> dict:
    from lunespy.client.transactions.constants import TransferType
    from lunespy.utils.crypto.converters import sign
    from lunespy.utils import lunes_to_unes
    from lunespy.utils import now
    from base58 import b58decode
    from struct import pack

    timestamp: int = transfer_data.get('timestamp', now())
    fee: int = transfer_data.get('fee', TransferType.fee.value)
    asset_id: str = transfer_data.get('asset_id', "")
    asset_fee: str = transfer_data.get('asset_fee', "")
    if asset_id == "":
        amount: int = lunes_to_unes(transfer_data['amount'])
    else:
        amount: int = transfer_data['amount']

    bytes_data: bytes = TransferType.to_byte.value + \
        b58decode(sender.public_key) + \
        (b'\1' + b58decode(asset_id) if asset_id != "" else b'\0') + \
        (b'\1' + b58decode(asset_fee) if asset_fee != "" else b'\0') + \
        pack(">Q", timestamp) + \
        pack(">Q", amount) + \
        pack(">Q", fee) + \
        b58decode(receiver.address)
    signature: bytes = sign(sender.private_key, bytes_data)
    return {
        "type": TransferType.to_int.value,
        "senderPublicKey": sender.public_key,
        "signature": signature.decode(),
        "timestamp": timestamp,
        "fee": fee,

        "recipient": receiver.address,
        "feeAsset": asset_fee,
        "assetId": asset_id,
        "amount": amount,
    }


def validate_transfer(sender: Account, receiver: Account, transfer_data: dict) -> bool:
    from lunespy.client.wallet.validators import validate_address
    from lunespy.utils import bcolors

    amount: int = transfer_data.get('amount', -1)

    if not sender.private_key:
        print(bcolors.FAIL + 'Sender `Account` not have a private key' + bcolors.ENDC)
        return False
    elif not validate_address(receiver.address, sender.network_id):
        return False
    elif amount <= 0:
        print(bcolors.FAIL + 'You dont pass `amount` param' + bcolors.ENDC)
        return False
    return True


# todo async
def send_transfer(mount_tx: dict, node_url: str) -> dict:
    from requests import post

    response = post(
        f'{node_url}/transactions/broadcast',
        json=mount_tx,
        headers={
            'content-type':
            'application/json'
        })
    
    if response.status_code in range(200, 300):
        mount_tx.update({
            'send': True,
            'response': response.json()
        })
        return mount_tx
    else:
        mount_tx.update({
            'send': False,
            'response': response.text
        })
        return mount_tx
